/*
 *	linux/fs/rbf/dir.c
 *	RBF File System - directory handling functions
 *
 *	Copyright (C) 2001 Andrew Cannon <ajc@gmx.net>
 *
 * from
 *  
 *  linux/fs/ext2/dir.c
 *  
 * Copyright (C) 1992, 1993, 1994, 1995
 * Remy Card (card@masi.ibp.fr)
 * Laboratoire MASI - Institut Blaise Pascal
 * Universite Pierre et Marie Curie (Paris VI)
 *
 * from
 *
 *  linux/fs/minix/dir.c
 *
 * Copyright (C) 1991, 1992  Linus Torvalds
 *
 * 	
 *  This file is subject to the terms and conditions of the
 *  GNU General Public License. See the file COPYING in the
 *  main directory of this archive for more details.
 */

#include "rbf_defs.h"
#include "rbf_block.h"

#undef DEBUG
#define DEBUG

#ifndef DEBUG
#undef DPRINTK
#define DPRINTK(x...)
#endif

static int rbf_readdir(struct file *f, void *dirent, filldir_t filldir)
{
	struct inode *dir = f->f_dentry->d_inode;
	struct super_block *s = dir->i_sb;
	struct buffer_head *bh;
	struct rbf_dirent *de;
	unsigned int offset;
	char name[RBF_NAMELEN];

	DPRINTK("in pos=%d\n", (int)f->f_pos);
	
	if (f->f_pos & (RBF_DIRENT_SIZE-1)) {
		printk(KERN_ERR "Bad f_pos=%08lx for %s:%08lx\n", (unsigned long)f->f_pos, 
			s->s_id, dir->i_ino);
		return -EBADF;
	}

	lock_kernel();

	while (f->f_pos < dir->i_size) {
		bh = rbf_logical_bread(dir, f->f_pos, &offset);
		if (bh == 0) {
			f->f_pos += s_su_sectsize - offset;
			continue;
		}
		do {
			de = (struct rbf_dirent *)(bh->b_data + offset);
			if (de->dir_name[0]) {
				int size = rbf_readname(name, de->dir_name);
				ino_t addr = rbf_read(de->dir_addr, 3);
				
				DPRINTK("entry '%.28s' addr %08lx\n", name, addr);
				if (filldir(dirent, name, size, f->f_pos, addr, DT_UNKNOWN) < 0) {
					brelse(bh);
					unlock_kernel();
					return 0;
				}
			}
			offset += RBF_DIRENT_SIZE;
			f->f_pos += RBF_DIRENT_SIZE;
		}
		while (offset < s_su_sectsize && f->f_pos < dir->i_size);
		brelse(bh);
	}

	DPRINTK("out\n");

	unlock_kernel();

	return 0;	
}

struct file_operations rbf_block_dir_operations = {
	.read =		generic_read_dir,
	.readdir =	rbf_readdir,
//	.ioctl =	rbf_ioctl,
//	.fsync =	rbf_sync_file,
	.fsync =	file_fsync
};



/*
 *	rbf_find_entry()
 *
 * finds an entry in the specified directory with the wanted name. It
 * returns the buffer in which the entry was found, and the entry itself.
 * Entry is guaranteed to be valid.
 */
struct rbf_dirent *rbf_find_entry(struct inode *dir, struct dentry *dentry, struct buffer_head **res_bh)
{
	struct super_block *s = dir->i_sb;
	const char *name = dentry->d_name.name;
	int namelen = dentry->d_name.len;
	unsigned long pos;
	int offset;
	struct buffer_head *bh;
	struct rbf_dirent *de;

	DPRINTK("name=%s, namelen=%d\n", name, namelen);
	
	if (namelen > RBF_NAMELEN)
		return NULL;
	
	pos = 0;
	while (pos < dir->i_size) {
		bh = rbf_logical_bread(dir, pos, &offset);
		if (bh == 0) {
			pos += s_su_sectsize - offset;
			continue;
		}
		do {
			de = (struct rbf_dirent *)(bh->b_data + offset);
			if (de->dir_name[0] && rbf_match(s, namelen, name, de->dir_name)) {
				*res_bh = bh;
				DPRINTK("out entry %ld\n", pos / RBF_DIRENT_SIZE);
				return de;
			}
			offset += RBF_DIRENT_SIZE;
			pos += RBF_DIRENT_SIZE;
		}
		while (offset < s_su_sectsize && pos < dir->i_size);
		brelse(bh);
	}

	DPRINTK("out 0\n");
	*res_bh = 0;
	return 0;
}

struct rbf_dirent *rbf_dotdot(struct inode *dir, struct buffer_head **res_bh)
{
	struct dentry d;		/* dummy for name lookup */
	
	d.d_name.name = "..";
	d.d_name.len = 2;
	return rbf_find_entry(dir, &d, res_bh);
}

void rbf_release_entry(struct buffer_head *bh)
{
	brelse(bh);
}

ino_t rbf_inode_by_name(struct inode *dir, struct dentry *dentry)
{
	ino_t ino = 0;
	struct rbf_dirent *de;
	struct buffer_head *bh;
	
	de = rbf_find_entry(dir, dentry, &bh);
	if (de) {
		ino = rbf_read(de->dir_addr, 3);
		brelse(bh);
	}
	return ino;
}

/* Releases the buffer */
void rbf_set_link(struct inode *dir, struct rbf_dirent *de, struct buffer_head *bh, struct inode *inode)
{
	rbf_write(de->dir_addr, 3, inode->i_ino);
	rbf_update_buffer(IS_SYNC(dir), bh);
	brelse(bh);
	dir->i_mtime = dir->i_ctime = CURRENT_TIME;
	mark_inode_dirty(dir);
}

/*
 *	Parent is locked.
 */
int rbf_add_link(struct dentry *dentry, struct inode *inode)
{
	struct inode *dir = dentry->d_parent->d_inode;
	struct super_block *s = dir->i_sb;
	const char *name = dentry->d_name.name;
	int namelen = dentry->d_name.len;
	struct buffer_head *bh;
	struct rbf_dirent *de;
	unsigned long pos = 0;
	int offset;
	int err;

	DPRINTK("name=%s, namelen=%d\n", name, namelen);

	err = rbf_validate_name(s, name, namelen);
	if (err)
		return err;

	while (pos < dir->i_size) {
		bh = rbf_logical_bread(dir, pos, &offset);
		if (bh == 0) {
			pos += s_su_sectsize - offset;
			continue;
		}
		do {
			de = (struct rbf_dirent *)(bh->b_data + offset);
			if (de->dir_name[0] == 0)
				goto got_it;
			err = -EEXIST;
			if (rbf_match(s, namelen, name, de->dir_name))	/* is this necessary? */
				goto out_bh;
			offset += RBF_DIRENT_SIZE;
			pos += RBF_DIRENT_SIZE;
		}
		while (offset < s_su_sectsize && pos < dir->i_size);
		brelse(bh);
	}
	/* reached end of dir - add a new entry */
	
	err = rbf_find_sector(dir, pos >> s_su_sectsize_bits, 0, 1);
	if (err < 0)
		goto out;
	bh = rbf_logical_bread(dir, pos, &offset);
	err = -EIO;
	if (bh == 0)
		goto out;
	de = (struct rbf_dirent *)(bh->b_data + offset);
	dir->i_size += RBF_DIRENT_SIZE;				/* expand directory */

got_it:
	DPRINTK("free entry %ld bh %08lx offset %d\n", pos / RBF_DIRENT_SIZE, (long int)bh, offset);
	
	strncpy(de->dir_name, name, RBF_NAMELEN + 1);		/* clear dir_res1 as well */
	de->dir_name[namelen - 1] |= 0x80;
	rbf_write(de->dir_addr, 3, inode->i_ino);
	rbf_update_buffer(IS_SYNC(dir), bh);
	brelse(bh);
	dir->i_mtime = dir->i_ctime = CURRENT_TIME;
	mark_inode_dirty(dir);
	DPRINTK("out\n");
	return 0;

out_bh:
	brelse(bh);
out:
	DPRINTK("out err=%d\n", err);
	return err;
}



/*
 * rbf_delete_entry deletes a directory entry.
 * Page is up-to-date. Releases the page.
 * 
 * ...should we shrink directory here if possible?
 */
int rbf_delete_entry(struct inode *dir, struct rbf_dirent *de, struct buffer_head *bh)
{
	de->dir_name[0] = 0;
	rbf_update_buffer(IS_SYNC(dir), bh);
	brelse(bh);
	dir->i_mtime = dir->i_ctime = CURRENT_TIME;
	mark_inode_dirty(dir);
	return 0;
}

/*
 * Set the first fragment of directory.
 */
int rbf_make_empty(struct inode *dir, struct inode *parent)
{
	struct buffer_head *bh;
	struct rbf_dirent *de;
	unsigned len = RBF_DIRENT_SIZE * 2;
	int offset;

	rbf_find_sector(dir, 0, 0, 1);		// expand file if necessary
	bh = rbf_logical_bread(dir, 0, &offset);
	if (bh == 0)
		return -ENOMEM;
	
	de = (struct rbf_dirent *)(bh->b_data + offset);
	memset(de, 0, len);
	
	de[0].dir_name[0] = '.';
	de[0].dir_name[1] = '.' | 0x80;
	rbf_write(de[0].dir_addr, 3, parent->i_ino);
	
	de[1].dir_name[0] = '.' | 0x80;
	rbf_write(de[1].dir_addr, 3, dir->i_ino);
	
	rbf_update_buffer(IS_SYNC(dir), bh);
	brelse(bh);
	dir->i_size = len;
	mark_inode_dirty(dir);
	return 0;
}

/*
 * routine to check that the specified directory is empty (for rmdir)
 */
int rbf_empty_dir(struct inode *dir)
{
	struct super_block *s = dir->i_sb;
	struct buffer_head *bh;
	struct rbf_dirent *de;
	unsigned pos = RBF_DIRENT_SIZE * 2;
	unsigned int offset;

	DPRINTK("in\n");
	
	while (pos < dir->i_size) {
		offset = pos & (s_su_sectsize-1);
		bh = rbf_logical_bread(dir, pos, &offset);
		if (bh == 0) {
			pos += s_su_sectsize - offset;
			continue;
		}
		do {
			de = (struct rbf_dirent *)(bh->b_data + offset);
			if (de->dir_name[0])
				goto not_empty;
			offset += RBF_DIRENT_SIZE;
			pos += RBF_DIRENT_SIZE;
		}
		while (offset < s_su_sectsize && pos < dir->i_size);
		brelse(bh);
	}
	DPRINTK("out 1\n");
	return 1;

not_empty:
	brelse(bh);
	DPRINTK("out 0\n");
	return 0;
}
