/*
 *	linux/fs/rbf/inode.c
 *	RBF File System - inode operations.
 *	
 *	Copyright (C) 2001 Andrew Cannon <ajc@gmx.net>
 * 	
 *  This file is subject to the terms and conditions of the
 *  GNU General Public License. See the file COPYING in the
 *  main directory of this archive for more details.
 */

#include "rbf_defs.h"

#undef DEBUG
#define DEBUG

#ifndef DEBUG
#undef DPRINTK
#define DPRINTK(x...)
#endif

static void rbf_truncate(struct inode *inode)
{
	struct super_block *s = inode->i_sb;
	long nsect = (inode->i_size + s_su_sectsize - 1) >> s_su_sectsize_bits;
	
	DPRINTK("size=%08x\n", (unsigned)inode->i_size);
	
	lock_kernel();
	if (nsect >= i_iu_nsect)	// NB: iu_nsect includes 1 for FD
		rbf_find_sector(inode, nsect - 1, 0, 1);
	else if (s_su_dealloc)
		rbf_contract(inode, nsect);
	unlock_kernel();
}

static int rbf_notify_change(struct dentry *dentry, struct iattr *attr)
{
	struct inode *inode = dentry->d_inode;
	struct super_block *s = inode->i_sb;
	int error;

	DPRINTK("in\n");

	error = inode_change_ok(inode, attr);
	if (error && !s_su_quiet)
		return error;
	error = inode_setattr(inode, attr);

	/* combine group & other permissions */
	inode->i_mode |= (inode->i_mode << 3 & 070) | (inode->i_mode >> 3 & 7);
	inode->i_mode &= ~s_su_umask;

	DPRINTK("out\n");
	return 0;
}

struct inode_operations rbf_file_inode_ops = {
	truncate:	rbf_truncate,
	setattr:	rbf_notify_change,
};


/*
 * Called when an inode is released. Note that this is different
 * from rbf_open_file: open gets called at every open, but release
 * gets called only when /all/ the files are closed.
 */
int rbf_release_file(struct inode *inode, struct file *filp)
{
	if (filp->f_mode & FMODE_WRITE) {
		/*
		 *	Release clusters allocated beyond end-of-file
		 *	if at end of file or "always dealloc" option set
		 */
		struct super_block *s = inode->i_sb;
		
		if (s_su_dealloc || (filp->f_pos == inode->i_size))
			return rbf_contract(inode, (inode->i_size + s_su_sectsize - 1) >> s_su_sectsize_bits);
	}
	return 0;
}



/*
 * Called by page-based I/O primitives
 */
int rbf_get_block(struct inode *inode, sector_t block, struct buffer_head *bh_result, int create)
{
	struct super_block *s = inode->i_sb;
	sector_t sector = block >> s_su_sect_shift;
	long phys = rbf_find_sector(inode, sector, 0, create);
	unsigned long end = (block + 1) << s->s_blocksize_bits;
	
	if (phys <= 0)
		return phys;
	
	map_bh(bh_result, inode->i_sb, phys);
	bh_result->b_blocknr = (phys << s_su_sect_shift) + (block & s_su_sect_mask);
	
	/*
	 *	mmu_private always points to the end of the last block accessed
	 *	in the file, or i_size initially. This is the point in the file at
	 *	which cont_prepare_write() will start zeroing when it extends the file.
	 */
	if (end > i_mmu_private) 
		i_mmu_private = end;
	
	return 0;
}


long rbf_find_sector(struct inode *inode, sector_t sector, int *lengthp, int create)
{
	struct super_block *s = inode->i_sb;
	struct buffer_head *bh;
	struct rbf_segment *sgp, *sge;
	sector_t segaddr;
	long segstart, segend, seglen = 0;
	long initial_size, nsect, alloc;
	long len, phys;
	int offset;
	int err = 0;
	
//	DPRINTK("in sector=%ld\n", sector);
	
	/*
	 *	Check whether sector is within cached current segment
	 */
	if (sector >= i_iu_segstart && sector < i_iu_segend) {
		phys = i_iu_segaddr + sector - i_iu_segstart;
		goto found;
	}
	
	/*
	 *	No... have to scan segment list.
	 *	Read it in.
	 */
	bh = rbf_bread(s, inode_adrs(s, inode->i_ino), &offset);
	if (bh == 0) {
		printk(KERN_ERR "RBF inode %s:%08lx error reading segment list\n", s->s_id, inode->i_ino);
		return -EIO;
	}
	sgp = (struct rbf_segment *)(bh->b_data + offset + sizeof(struct rbf_inode));
	sge = sgp + s_su_maxseg;
	
	if (sector >= i_iu_segend) {
		segaddr = i_iu_segaddr;	/* load current segment */
		segstart = i_iu_segstart;
		segend = i_iu_segend;
		seglen = segend - segstart;
		sgp += i_iu_segnum;	/* point to next segment */
	}
	else {
		segstart = segend = 0;		/* start at beginning of segment list */
		sgp--;
	}
	
	while (++sgp != sge) {
		phys = rbf_read(sgp->sg_addr, 3);
		len = rbf_read(sgp->sg_size, 2);
		if (len == 0 || phys == 0)
			break;			/* end of allocated segments */
		segaddr = phys;
		segstart = segend;
		seglen = len;
		segend += seglen;
//		DPRINTK("addr=%08lx start=%08lx end=%08lx len=%08lx\n", segaddr, segstart, segend, seglen);
		if (sector < segend)
			goto setseg;
	}
	
	if (!create) {
		brelse(bh);
		return 0;			/* block not allocated */
	}
	
	/*
	 *	Expand File
	 *	
	 *	sgp -> next segment list entry
	 *	if segend!=0 segaddr/segstart/segend/seglen refer to previous entry
	 *	
	 *	ToDo: make this a separate function!
	 */
	initial_size = segend;
	
	while (segend <= sector) {
		/*
		 * try to extend last segment.
		 * repeat until unsuccessful since each extension will only
		 * allocate a maximum of one bitmap sector's worth of clusters.
		 */
		alloc = max((long) (sector + 1 - segend), (long) s_su_minalloc);
		if (segend && (nsect = min(alloc, (long) (65536 - s_su_clus - seglen)))
			   && (nsect = rbf_extend_allocation(s, segaddr + seglen, nsect))) {
			if (nsect < 0) {
				err = nsect;
				break;
			}
			segend += nsect;
			seglen += nsect;
			rbf_write(sgp[-1].sg_size, 2, seglen);
			mark_buffer_dirty(bh);
			i_iu_nsect += nsect;
			inode->i_blocks = sect_iblock(s, i_iu_nsect);
			continue;
		}
		/*
		 * could not extend - add new segment
		 */
		if (sgp == sge) {
			err = -EFBIG;
			break;		/* segment list full */
		}
		seglen = min(alloc, 65536L - s_su_clus);
		if ((seglen = rbf_allocate(s, &segaddr, seglen))) {
			if (seglen < 0) {
				err = seglen;
				break;
			}
			segstart = segend;
			segend += seglen;
			rbf_write(sgp->sg_addr, 3, segaddr);
			rbf_write(sgp->sg_size, 2, seglen);
			mark_buffer_dirty(bh);
			i_iu_nsect += seglen;
			inode->i_blocks = sect_iblock(s, i_iu_nsect);
			sgp++;
			continue;
		}	
		err = -ENOSPC;
		break;
	}
	
	if (sector >= segend) {
		/*
		 *	Fail. Return newly allocated sectors
		 */
		rbf_contract(inode, initial_size);
		brelse(bh);
		return err;
	}
	sgp--;
	
setseg:
	/*
	 *	Cache current seg for next call to rbf_find_sector()
	 */
	i_iu_segnum = sgp - (struct rbf_segment *)(bh->b_data + offset + sizeof(struct rbf_inode));
	i_iu_segaddr = segaddr;
	i_iu_segstart = segstart;
	i_iu_segend = segend;
	phys = segaddr + sector - segstart;
	brelse(bh);

found:
//	DPRINTK("out phys=%ld\n", phys);
	if (lengthp)
		*lengthp = i_iu_segend - sector;
	return phys;
}


/*
 *	Reduce file allocation size to <sector> sectors
 */
int rbf_contract(struct inode *inode, sector_t sector)
{
	struct super_block *s = inode->i_sb;
	struct buffer_head *bh;
	struct rbf_segment *sgp, *sge;
	long len, phys, segaddr = 0, segfinal, segstart = 0, segend = 0, seglen = 0;
	int nseg = 0;
	int offset;
	
	DPRINTK("in sector=%d\n", (int) sector);
	
	if ((((sector - 1) ^ (i_iu_nsect - 2)) & -s_su_clus) == 0)	/* no change */
		return 0;
	
	/*
	 *	Read segment list.
	 */
	bh = rbf_bread(s, inode_adrs(s, inode->i_ino), &offset);
	if (bh == 0) {
		printk(KERN_ERR "RBF inode %s:%08lx error reading segment list\n", s->s_id, inode->i_ino);
		return -EIO;
	}
	sgp = (struct rbf_segment *)(bh->b_data + offset + sizeof(struct rbf_inode));
	sge = sgp + s_su_maxseg;
	
	if (sector == 0) {
		segaddr = rbf_read(sgp->sg_addr, 3);
		seglen = rbf_read(sgp->sg_size, 2);
		if (seglen == 0 || segaddr == 0)
			goto out;
		segend = seglen;
	}
	else if (i_iu_segnum != -1 && sector >= i_iu_segstart) {
		segaddr = i_iu_segaddr;	/* load current segment */
		segstart = i_iu_segstart;
		segend = i_iu_segend;
		seglen = segend - segstart;
		sgp += i_iu_segnum;
	}
	else sgp--;				/* start at beginning of segment list */
	
	while (sector > segend) {
		if (++sgp == sge)
			goto out;
		phys = rbf_read(sgp->sg_addr, 3);
		len = rbf_read(sgp->sg_size, 2);
		if (len == 0 || phys == 0)
			goto out;
		segaddr = phys;
		segstart = segend;
		seglen = len;
		segend += seglen;
	}
	
	/*
	 *	determine new endpoint of final segment
	 *	round up to cluster boundary
	 */
	segfinal = (segaddr + sector - segstart + s_su_clus - 1) & -s_su_clus; 
	
	while (++sgp != sge) {
		phys = rbf_read(sgp->sg_addr, 3);
		len = rbf_read(sgp->sg_size, 2);
		if (len == 0 || phys == 0)
			break;				/* end of allocated segments */
		segaddr = phys;
		seglen = len;
		nseg++;
	}
	
	/*
	 *	deallocate whole segments
	 */
	--sgp;
	while (nseg--) {
		rbf_write(sgp->sg_size, 2, 0);		/* delete segment */
		rbf_deallocate(s, segaddr, seglen);
		i_iu_nsect -= seglen;
		--sgp;
		segaddr = rbf_read(sgp->sg_addr, 3);
		seglen = rbf_read(sgp->sg_size, 2);
	}
	
	/*
	 *	truncate final segment
	 */
	len = seglen;
	seglen = segfinal - segaddr;		/* new length */
	len -= seglen;				/* amount deallocated */
	rbf_write(sgp->sg_size, 2, seglen);
	rbf_deallocate(s, segfinal, len);
	
	mark_buffer_dirty(bh);
	
	/*
	 *	update segment cache
	 */
	segend -= len;
	i_iu_segnum = sgp - (struct rbf_segment *)(bh->b_data + offset + sizeof(struct rbf_inode));
	i_iu_segaddr = segaddr;
	i_iu_segstart = segstart;
	i_iu_segend = segend;
	if (segend == 0)
		i_iu_segnum = -1;
	
	i_iu_nsect -= len;
	inode->i_blocks = sect_iblock(s, i_iu_nsect);
	
out:
	brelse(bh);
	return 0;
}


/*
 *	Determine number of sectors allocated to inode
 *	
 *	Called by rbf_read_inode() - also does a few integrity checks
 */
long rbf_scan_fd(struct inode *inode, struct rbf_inode *ip)
{
	struct super_block *s = inode->i_sb;
	struct rbf_segment *sgp, *sge;
	int len, addr, nsect = 1;		/* count initial sector for FD */
	
	DPRINTK("in\n");
	
	/*
	 *	Get segment list.
	 */
	sgp = (struct rbf_segment *)(ip + 1);
	sge = sgp + s_su_maxseg;
	
	/*
	 *	Quick sanity check on inode
	 */
	addr = rbf_read(sgp->sg_addr, 3);
	len = rbf_read(sgp->sg_size, 2);
	if (s_su_clus_bits && addr != inode_adrs(s, inode->i_ino) + 1)
		goto bad;
	
	if (len) {
		if (addr == 0)
			goto bad;
		nsect += len;
		while (++sgp != sge) {
			addr = rbf_read(sgp->sg_addr, 3);
			len = rbf_read(sgp->sg_size, 2);
			if (len == 0)
				break;			/* end of allocated segments */
			if (addr == 0 || addr >= s_su_nsect)
				goto bad;
			if ((addr|len) & (s_su_clus - 1))
				goto bad;
			nsect += len;
		}
	}
	
	/*
	 *	If sector size > blocksize then check that no segments are allocated
	 *	past the first BLOCK in FD, since rbf_get_block() currently doesn't
	 *	handle these at all. What to do ...???
	 *	
	 *	Print a warning & return 0... rbf_read_inode() will bomb out.
	 */
	if (sgp == sge && s_su_sectsize > s->s_blocksize) {
		/*
		 *	NB: this assumes blocksize of 4096, which conveniently results
		 *	in the first block ending exactly on a segment boundary
		 */
		struct buffer_head *bh = sb_bread(s, inode_block(s, inode->i_ino) + 1);
		
		if (bh && rbf_read(((struct rbf_segment *)bh->b_data)->sg_size, 2)) {
			if (!s_su_quiet)
				printk(KERN_ERR "RBF inode %s:%08lx contains too many segments\n",
						s->s_id, inode->i_ino);
			nsect = 0;
		}
		brelse(bh);
	}

	DPRINTK("out nsect %d\n", nsect);
	i_iu_nsect = nsect;
	return sect_iblock(s, nsect);

bad:
	DPRINTK("out nsect 0\n");
	if (!s_su_quiet)
		printk(KERN_ERR "RBF invalid inode %s:%08lx\n", s->s_id, inode->i_ino);
	i_iu_nsect = 0;
	return 0;
}


/*
 * Allocate new inode on disk
 */
struct inode *rbf_new_inode(const struct inode *dir, int mode)
{
	struct super_block *s = dir->i_sb;
	struct inode *inode;
	struct buffer_head *bh;
	int segsize;
	int offset;
	long ino;
	int i, err;

	DPRINTK("in\n");
	
	/* Cannot create files in a deleted directory */
	if (!dir || !dir->i_nlink)
		return ERR_PTR(-EPERM);

	inode = new_inode(s);
	if (!inode)
		return ERR_PTR(-ENOMEM);

	err = -ENOSPC;
	if ((segsize = rbf_allocate(s, (sector_t *) &ino, 1)) == 0)		/* allocate FD sector */
		goto fail;
	
	/*
	 *	clear segment list
	 */
	bh = rbf_bread(s, inode_adrs(s, ino), &offset);
	if (bh == 0) {
		printk(KERN_ERR "RBF inode %s:%08lx error reading segment list\n", s->s_id, ino);
		rbf_deallocate(s, ino, 1);
		err = -EIO;
		goto fail;
	}
	memset(bh->b_data + offset, 0, min(s_su_sectsize, (long)s->s_blocksize));
	if (segsize > 1) {
		/*
		 *	generate initial segment
		 */
		struct rbf_segment *sgp;
		
		sgp = (struct rbf_segment *)(bh->b_data + offset + sizeof(struct rbf_inode));
		rbf_write(sgp->sg_addr, 3, ino + 1);
		rbf_write(sgp->sg_size, 2, segsize - 1);
	}
	mark_buffer_dirty(bh);
	brelse(bh);
	/*
	 *	if sector size > blocksize then zero remainder of FD sector
	 */
	for (i = 1; i <= s_su_sect_mask; i++) {
		bh = sb_bread(s, inode_block(s, ino) + i);
		if (bh == 0)
			continue;
		memset(bh->b_data, 0, s->s_blocksize);
		mark_buffer_dirty(bh);
		brelse(bh);
	}
	i_iu_segnum = -1;
	i_iu_segend = 0;
	
	inode->i_uid = current->fsuid;
	inode->i_gid = current->fsgid;
	
	inode->i_mode = mode;
	inode->i_ino = ino;
	i_iu_nsect = segsize;
	i_mmu_private = 0;
	inode->i_blocks = sect_iblock(s, i_iu_nsect);
	inode->i_mtime.tv_sec = inode->i_atime.tv_sec = inode->i_ctime.tv_sec = i_iu_create = CURRENT_TIME.tv_sec;
	inode->i_mtime.tv_nsec = inode->i_atime.tv_nsec = inode->i_ctime.tv_nsec = 0;
	insert_inode_hash(inode);
	inode->i_generation++;
	mark_inode_dirty(inode);

	DPRINTK("out %08lx\n", ino);
	return inode;

fail:
	make_bad_inode(inode);
	iput(inode);
	DPRINTK("out err=%d\n", err);
	return ERR_PTR(err);
}
