/*
 *	linux/fs/rbf/rbf_bitops.h
 *	RBF File System - bit operations.
 *	
 *	Copyright (C) 2001 Andrew Cannon <ajc@gmx.net>
 *	
 * from
 * 
 *  linux/include/asm-m68k/bitops.h
 * 	
 *  This file is subject to the terms and conditions of the
 *  GNU General Public License. See the file COPYING in the
 *  main directory of this archive for more details.
 */


/*
 *  Basic bit-scanning operation for RBF bitmaps (BE byte/BE bit ordering)
 *  
 *  return offset from MSB end to first set bit in BE 32 bit word
 *  return 32 if no bit set.
 */
#if defined(__mc68000__)

#undef RBF_HIGHZ_TABLE
static inline int rbf_count_zero_msbs(u32 n)
{
	int res;
	
	__asm__ __volatile__ ("bfffo %1{#0,#0},%0" : "=d" (res) : "d" (n));
	return res;
}
#else

extern char rbf_highz[];
#define RBF_HIGHZ_TABLE

static inline int rbf_count_zero_msbs(u32 n)
{
	if (n & 0xffff0000)
		if (n & 0xff000000)
			return rbf_highz[n >> 24];
		else	return rbf_highz[n >> 16] + 8;
	else if (n & 0xffff)
		if (n & 0xff00)
			return rbf_highz[n >> 8] + 16;
		else	return rbf_highz[n] + 24;
	else return 32;
}
#endif



static inline void rbf_set_bits(void *vaddr, int offset, int length)
{
	u32 *p = (u32 *)vaddr + (offset >> 5);
	u32 allones = (u32) ~0UL;
	int bit = offset & 31;

	if (bit) {			/* set bits in first partial word */	
		int w = min(length, 32 - bit);
	
		*p++ |= cpu_to_be32((allones << (32 - w)) >> bit);
		length -= w;
	}
	while (length >= 32) {		/* set whole words */
		*p++ = allones;
		length -= 32;
	}
	if (length)			/* set bits in last partial word */
		*p |= cpu_to_be32(allones << (32 - length));
}

static inline void rbf_clear_bits(void *vaddr, int offset, int length)
{
	u32 *p = (u32 *)vaddr + (offset >> 5);
	int bit = offset & 31;
	u32 allones = (u32) ~0UL;		

	if (bit) {			/* clear bits in first partial word */	
		int w = min(length, 32 - bit);
	
		*p++ &= cpu_to_be32(~((allones << (32 - w)) >> bit));
		length -= w;
	}
	while (length >= 32) {		/* clear whole words */
		*p++ = 0;
		length -= 32;
	}
	if (length)			/* clear bits in last partial word */
		*p &= cpu_to_be32(allones >> length);
}

static inline int rbf_find_first_zero_bit(void *vaddr, int size)
{
	u32 *p = vaddr, *addr = vaddr;
	u32 allones = (u32) ~0UL;
	int res, n;

	if (size <= 0)
		return 0;

	n = (size + 31) >> 5;
	while (*p++ == allones)
		if (--n == 0)
			return 32 * (p - addr) - (-size & 31);

	res = rbf_count_zero_msbs(~be32_to_cpu(*--p));
	
	return min((int)(((p - addr) << 5) + res), size);
}

static inline int rbf_find_next_zero_bit(void *vaddr, int size, int offset)
{
	u32 *addr = vaddr;
	u32 *p = addr + (offset >> 5);
	int clr = 0, bit = offset & 31, res;

	if (offset >= size)
		return size;

	if (bit) {
		/* Look for zero in first longword */
		u32 allones = (u32) ~0UL;		
		res = rbf_count_zero_msbs(~be32_to_cpu(*p) & (allones >> bit));

		if (res < 32)
			return min((offset & ~31) + res, size);
                clr = 32 - bit;
		p++;
	}
	
	/* No zero yet, search remaining full bytes for a zero */
	
	return offset + clr + rbf_find_first_zero_bit(p, size - 32 * (p - addr));
}

static inline int rbf_find_first_one_bit(void *vaddr, int size)
{
	u32 *p = vaddr, *addr = vaddr;
	int res, n;

	if (size <= 0)
		return 0;

	n = (size + 31) >> 5;
	while (*p++ == 0)
		if (--n == 0)
			return 32 * (p - addr) - (-size & 31);

	res = rbf_count_zero_msbs(be32_to_cpu(*--p));
	
	return min((int)(((p - addr) << 5) + res), size);
}

static inline int rbf_find_next_one_bit(void *vaddr, int size, int offset)
{
	u32 *addr = vaddr;
	u32 *p = addr + (offset >> 5);
	int set = 0, bit = offset & 31, res;

	if (offset >= size)
		return size;

	if (bit) {
		/* Look for one in first longword */
		u32 allones = (u32) ~0UL;
		res = rbf_count_zero_msbs(be32_to_cpu(*p) & (allones >> bit));
		if (res < 32)
			return min((offset & ~31) + res, size);
                set = 32 - bit;
		p++;
	}
	
	/* No one yet, search remaining full bytes for a one */
	
	return offset + set + rbf_find_first_one_bit(p, size - 32 * (p - addr));
}
